from ship import Ship
from shot import PlasmaBlast
from OpenGL.GL import *
from OpenGL.GLU import *
from data import filepath
from vector import Vector
import random
import effects

models = {}

class Enemy(Ship):
    speed = 0.0
    scoreValue = 100
    maxLife = 2.0
    scale = 0.1
    def loadModel(self, filename):
        if models.get(filename, None):
            self.points = models[filename]['points']
            self.normals = models[filename]['normals']
        else:
            f = open (filepath(filename))
            lineNum = 0
            self.normals = []
            self.points = []
            for line in f.readlines():
                coords = [float(x.strip()) for x in line.split()]
                if lineNum % 4 == 0:
                    self.normals.append (coords)
                else:
                    self.points.append(coords)
                lineNum += 1
            models[filename] = {'points': self.points, 'normals': self.normals}
        self.draw = self.drawWithModel

    def loadTexturedModel(self, filename):
        if models.get(filename, None):
            self.normals = models[filename]['normals']
            self.points = models[filename]['points']
        else:
            f = open (filepath(filename))
            self.points = []
            self.normals = []
            lineNum = 0
            for line in f.readlines():
                coords = [float(x.strip()) for x in line.split()]
                if lineNum % 4 == 0:
                    self.normals.append (coords)
                else:
                    self.points.append(coords)
                lineNum += 1
            models[filename] = {'normals': self.normals, 'points': self.points}
        self.draw = self.drawWithTexture

    def loadMaterialModel(self, filename):
        if models.get(filename, None):
            self.objs = models[filename]['objs']
            self.materials = models[filename]['materials']
        else:
            f = open (filepath(filename))
            self.objs = []
            self.materials = {}
            line = f.readline()
            while line.startswith ('Material'):
                parts = [x.strip() for x in line.split()]
                self.materials[parts[1]] = parts[2:]
                line = f.readline()
            while line != 'End\n':
                material = line[4:]
                lineNum = 0
                obj = {'material': material, 'points': [], 'normals': []}
                line = f.readline()
                while not line.startswith('Use ') and not line == 'End\n':
                    coords = [float(x.strip()) for x in line.split()]
                    if lineNum % 4 == 0:
                        obj['normals'].append (coords)
                    else:
                        obj['points'].append(coords)
                    lineNum += 1
                    line = f.readline()
            models[filename] = {'objs': self.objs, 'materials': self.materials}
        self.draw = self.drawWithMaterial

    def drawWithModel (self):
        glPushMatrix()
        glTranslatef(self.pos.x, self.pos.y, self.pos.z)
        glScalef(self.scale, self.scale, self.scale)
        m = list(self.section.right.coords()) + [0] + list(self.section.up.coords()) + [0] + list(self.section.forward.coords()) + [0, 0, 0, 0, 1]
        glMultMatrixf(m)
        triang = 0
        vert = 0
        glDisable(GL_TEXTURE_2D)
        glBegin(GL_TRIANGLES)
        glColor3fv((1, 1, 1))
        for x, y, z in self.points:
            if vert == 0:
                glNormal3f (self.normals[triang][0], self.normals[triang][1], self.normals[triang][2])
            glVertex3f(x, y, z)
            vert += 1
            if vert == 3:
                vert = 0
                triang += 1
        glEnd()
        glEnable(GL_TEXTURE_2D)
        glPopMatrix()
        # Burning effect
        if self.life < self.maxLife / 3:
            effects.burn(self.pos)

    def drawWithTexture (self):
        glBindTexture(GL_TEXTURE_2D, self.texture.texid)
        glPushMatrix()
        glTranslatef(self.pos.x, self.pos.y, self.pos.z)
        m = list(self.section.right.coords()) + [0] + list(self.section.up.coords()) + [0] + list(self.section.forward.coords()) + [0, 0, 0, 0, 1]
        glMultMatrixf(m)
        if self.rot.x:
            glRotatef(self.rot.x, 1, 0, 0)
        if self.rot.y:
            glRotatef(self.rot.y, 0, 1, 0)
        if self.rot.z:
            glRotatef(self.rot.z, 0, 0, 1)
        glScalef(self.scale, self.scale, self.scale)
        triang = 0
        vert = 0
        glEnable(GL_CULL_FACE)
        glCullFace(GL_BACK)
        glEnable(GL_NORMALIZE)
        glBegin(GL_TRIANGLES)
        glColor3fv((1, 1, 1))
        for x, y, z, u, v in self.points:
            ns = self.normals[triang]
            if vert == 0:
                glNormal3f (*ns)
            glTexCoord2f(u, v)
            glVertex3f(x, y, z)
            vert += 1
            if vert == 3:
                vert = 0
                triang += 1
        glEnd()
        glDisable(GL_NORMALIZE)
        glDisable(GL_CULL_FACE)
        glPopMatrix()
        # Burning effect
        if self.life < self.maxLife / 3:
            effects.burn(self.pos)

    def __init__ (self, section, px, py, texture, world):
        Ship.__init__ (self, section)
        self.px = px
        self.py = py
        self.strafe (px, py)
        self.texture = texture
        self.world = world
        self.rot = Vector(0,0,0)
        self.dead = False
        self.life = self.maxLife

    def aimAndFire(self):
        """Shoot a plasma bolt towards the player"""
        aim = self.world.hero.pos + self.world.hero.section.path_vector
        b = PlasmaBlast(self.pos, (aim - self.pos).normalize(), self.section, 
                self.world, self.world.textures['blast'], False)
        effects.plasma ((self.pos -self.world.hero.pos).norm())
        b.world = self.world
        self.world.shots.append (b)

    def hit (self, damage):
        """Called when self is hit"""
        if damage=='collision':
            self.dead = True
        else:
            self.life -= damage
            self.dead = self.life <= 0
        if self.dead:
            effects.smallExplosion (self.pos, (self.pos - self.world.hero.pos).norm())


class Enemy1(Enemy):
    """Non shooting enemy, constant speed"""
    scoreValue = -200
    def __init__ (self, section, px, py, texture, world, speed=0.018):
        self.speed = speed
        Enemy.__init__ (self, section, px, py, texture, world)
        self.loadTexturedModel('models/alpha.rat')
        self.radius = 0.14
        if speed <= 0:
            self.rot.y = 180

    def move(self):
        self.forward (self.speed)

    def serialize(self, world):
        return "Enemy1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

SimpleEnemy = Enemy1

class Turret(Enemy):
    """Fixed position enemy"""
    scoreValue = 300
    speed = 0.00
    maxLife = 1.2
    scale = 0.2
    def __init__ (self, section, px, py, rot, texture, world):
        Enemy.__init__ (self, section, px, py, texture, world)
        self.loadTexturedModel('models/torreta.rat')
        self.gunTimer = random.randint(10, 30)
        self.radius = 0.16
        self.rot.z = rot

    def move(self):
        if self.gunTimer > 0:
            self.gunTimer -= 1
        else:
            self.aimAndFire()
            self.gunTimer = random.randint(20, 50)

    def serialize(self, world):
        return "Turret, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Tracker(Enemy):
    """This guy stands still until you get close. Then, it matches your speed"""
    keepAt = 2.5
    scoreValue = 250
    def __init__ (self, section, px, py, texture, world, reload = 0):
        Enemy.__init__ (self, section, px, py, texture, world)
        self.loadTexturedModel('models/beta.rat')
        self.reload = reload
        self.gunTimer = random.randint(10, 30) + self.reload
        self.active = False
        self.ventaja = -1
        self.radius = 0.12

    def move(self):
        if self.active:
            self.forward (self.world.hero.speed)
        else:
            if self.ventaja < 0:
                if self.section == self.world.hero.section.next:
                    self.ventaja = (self.section.pos - self.world.hero.section.pos).norm()
            else:
                self.ventaja -= self.world.hero.speed
                if self.ventaja < self.keepAt:
                    self.active = True
                    self.scoreValue=500

        #print self.speed == 0 and 'asleep' or 'active'
        if self.gunTimer > 0:
            self.gunTimer -= 1
        else:
            self.aimAndFire()
            self.gunTimer = random.randint(20, 50) + self.reload

    def serialize(self, world):
        return "Tracker, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Alpha(Enemy):
    def __init__ (self, section, texture, world, speed=0.018):
        self.speed = speed
        Enemy.__init__ (self, section, 0, 0, texture, world)
        self.loadTexturedModel('models/alpha.rat')
        self.rot.x = 90
        self.rot.y = 180
        self.radius = self.radius*2

    def move(self):
        #Showcase mode for checking models
        self.rot.x = (self.rot.x+1)%360
        self.rot.y = (self.rot.y+1)%360
        self.rot.z = (self.rot.z+1)%360

    def serialize(self, world):
        return "Beta1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Dodger(Enemy):
    """Vuela por el tunel moviendose por la seccion en forma aleatoria"""
    
    scoreValue = 200
    
    def __init__ (self, section, texture, world, speed=0.018, reload=0):
        self.speed = speed
        Enemy.__init__ (self, section, 0, 0, texture, world)
        self.loadTexturedModel('models/gamma.rat')
        #self.rot.x = 90
        #self.rot.y = 180
        self.reload = reload
        self.gunTimer = random.randint(10, 30) + self.reload
        self.radius = self.radius*2
        self.avance = 0
        self.posx = 0
        self.posy = 0
        self.objx = random.random()-0.5
        self.objy = random.random()-0.5

    def move(self):
        #Showcase mode for checking models
        #self.rot.x = (self.rot.x+1)%360
        #self.rot.y = (self.rot.y+1)%360
        #self.rot.z = (self.rot.z+1)%360
        self.forward(self.speed)
        vx = (self.objx - self.posx) / 60
        vy = (self.objy - self.posy) / 60
        self.strafe(vx, vy)
        self.avance += 1
        if self.avance == 60:
            self.avance = 0
            self.posx = self.objx
            self.posy = self.objy
            self.objx = random.random()-0.5
            self.objy = random.random()-0.5
        if self.gunTimer > 0:
            self.gunTimer -= 1
        else:
            self.aimAndFire()
            self.gunTimer = random.randint(10, 30) + self.reload

    def serialize(self, world):
        return "Dodger, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Gamma(Enemy):
    def __init__ (self, section, texture, world, speed=0.018):
        self.speed = speed
        Enemy.__init__ (self, section, 0, 0, texture, world)
        self.loadTexturedModel('models/gamma.rat')
        self.rot.x = 90
        self.rot.y = 180
        self.radius = self.radius*2

    def move(self):
        #Showcase mode for checking models
        self.rot.x = (self.rot.x+1)%360
        self.rot.y = (self.rot.y+1)%360
        self.rot.z = (self.rot.z+1)%360

    def serialize(self, world):
        return "Gamma1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Delta(Enemy):
    def __init__ (self, section, texture, world, speed=0.018):
        self.speed = speed
        Enemy.__init__ (self, section, 0, 0, texture, world)
        self.loadTexturedModel('models/delta.rat')
        self.rot.x = 90
        self.rot.y = 180
        self.radius = self.radius*2

    def move(self):
        #Showcase mode for checking models
        self.rot.x = (self.rot.x+1)%360
        self.rot.y = (self.rot.y+1)%360
        self.rot.z = (self.rot.z+1)%360

    def serialize(self, world):
        return "Delta1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Epsilon(Enemy):
    def __init__ (self, section, texture, world, speed=0.018):
        self.speed = speed
        Enemy.__init__ (self, section, 0, 0, texture, world)
        self.loadTexturedModel('models/epsilon.rat')
        self.rot.x = 90
        self.rot.y = 180
        self.radius = self.radius*2.5

    def move(self):
        #Showcase mode for checking models
        self.rot.x = (self.rot.x+1)%360
        self.rot.y = (self.rot.y+1)%360
        self.rot.z = (self.rot.z+1)%360

    def serialize(self, world):
        return "Epsilon1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class EnemyPasser(Enemy):
    """Invisible/untouchable until you get past it. Then, it appears,
    accelerates in front of you, and opens fire."""

    scoreValue = 500

    def __init__ (self, section, px, py, texture, world, reload = 0):
        self.speed = 0
        Enemy.__init__ (self, section, px, py, texture, world)
        #self.loadTexturedModel('models/senemy5.rat')
        self.loadModel('models/gamma.rat')
        self.reload = reload
        self.rot.x = 0
        self.rot.y = 0
        self.radius = -1
        self.visible = False
        self.mode = 0
        del self.draw

    def move(self):
        h = self.world.hero
        d = (self.pos - h.pos).norm()
        MATCH_ACCEL = 0.0005
        if self.mode == 0: # Hiding, waiting for hero to pass
            if d < 1 and (self.pos - h.pos).dot(self.section.forward) > 0:
                self.visible = True
                self.speed = h.burnerSpeed * 1.2 + random.random() * 0.1
                self.mode = 1
        elif self.mode == 1: # Passing hero
            self.forward (self.speed)
            if d > 0.5:
                self.radius = 0.2
            if d > 1.8:
                self.mode = 2
                self.gunTimer = 25
        else: # Matching speed and firing
            # Match speed, with some delay. This might allow us to get closer
            # or farther
            if abs(self.speed - h.speed) < MATCH_ACCEL:
                self.speed = h.speed
            elif self.speed > h.speed:
                self.speed -= MATCH_ACCEL
            else:
                self.speed += MATCH_ACCEL
            # Maximum speed: a bit slower than the player with burner
            self.speed = min (self.speed, h.burnerSpeed * 0.7 + h.baseSpeed*0.3)
            #effects.partEngine.addCloud (self.pos, 5, 0.05)
            self.forward (self.speed)
            self.gunTimer -= 1
            if self.gunTimer == 0:
                self.aimAndFire()
                self.gunTimer = 25 + self.reload

    def draw (self):
        if self.visible:
            self.drawWithTexture()
            #self.drawWithModel()

    def serialize(self, world):
        return "Epsilon1, %d, %f, %f\n" % (world.map.index (self.section), self.px, self.py)

class Freighter (Enemy):
    scoreValue = -1000
    scale = 0.2
    maxLife = 30.0
    defaultRadius = 0.5
    def __init__ (self, section, px, py, texture, world, speed = 0.008):
        Enemy.__init__ (self, section, px, py, texture, world)
        self.loadTexturedModel('models/carguero.rat')
        self.radius = self.defaultRadius
        self.speed = speed
        self.scale = self.scale * 2

    def move(self):
        self.forward (self.speed)
