from OpenGL.GL import *
from OpenGL.GLU import *
import pygame
from pygame.locals import *
import random
import os

from section import Section, LastSection
from vector import Vector
from texloader import LoadGLTextures, LoadGLTexturesAlpha
from hero import Hero
from shot import Shot, PlasmaBlast
from enemies import Enemy1, Tracker, Turret, Dodger, EnemyPasser, SimpleEnemy, Freighter
from enemies import Alpha, Gamma, Delta, Epsilon
from data import filepath
from statics import Score
import effects

FPS = 60
MOVE_SPEED = 0.01 # Effect of the arrow keys on the player

chars = {'0': [0, 0.085],
        '1': [0.09, 0.06],
        '2': [0.15, 0.09],
        '3': [0.24, 0.085],
        '4': [0.33, 0.09],
        '5': [0.42, 0.09],
        '6': [0.51, 0.085],
        '7': [0.59, 0.08],
        '8': [0.675, 0.085],
        '9': [0.77, 0.09],
        '-': [0.94, 0.05]}

class World(object):
    MAX_SECTIONS = 20 # Maximum number of sections drawn
    def __init__(self, fovy, width, height, level=None):
        self.quit = False
        self.won = False
        self.lost = False
        self.enemies = []
        self.shots = []
        self.scoreLabels = []
        self.map = [] # List of tunnel sections
        self.mousePos = (width / 2, height / 2)
        self.fovy = fovy # Field-of-view angle
        self.width = width
        self.height = height
        self.enemiesTextures = LoadGLTextures ({'alpha1': "textures/alpha1.bmp",
                                                'alpha2': "textures/alpha2.bmp",
                                                'beta1': "textures/beta1.bmp",
                                                'beta2': "textures/beta2.bmp",
                                                'gamma1': "textures/gamma1.bmp",
                                                'gamma2': "textures/gamma2.bmp",
                                                'delta1': "textures/delta1.bmp",
                                                'delta2': "textures/delta2.bmp",
                                                'epsilon1': "textures/epsilon1.bmp",
                                                'epsilon2': "textures/epsilon2.bmp",
                                                'carguero': "textures/Carguero.bmp",
                                                'torreta': "textures/torreta.bmp"
                                                })
        self.textures = LoadGLTextures ({'tunnel': "tunnel.bmp",
                                         'gauge': "gauge.bmp",
                                         'enemies': "enemies.bmp",
                                         'blast': "plasmablast.bmp",
                                         'powerblast': "orangeplasma.bmp"})
        self.alphaTextures = LoadGLTexturesAlpha ({'ready': "ready.png",
                                                   'crosshair': "crosshair2.png"})
        self.mapTextures = LoadGLTextures (
                    {'tunnel': "tunnel.bmp",
                    'b02': 'textures/b02.bmp',
                    'b03': 'textures/b03.bmp',
                    'b04': 'textures/b04.bmp',
                    'b06': 'textures/b06.bmp',
                    'b07': 'textures/b07.bmp',
                    'b19': 'textures/b19.bmp',
                    'envmap01': 'textures/envmap01.bmp',
                    'envmap04': 'textures/envmap04.bmp',
                    'envmap05': 'textures/envmap05.bmp',
                    'envmap11': 'textures/envmap11.bmp',
                    'envmap15': 'textures/envmap15.bmp',
                    'phong01': 'textures/phong01.bmp',
                    'spiral05': 'textures/spiral05.bmp',
                    'spiral07': 'textures/spiral07.bmp',
                    'spiral08': 'textures/spiral08.bmp',
                    'spiral09': 'textures/spiral09.bmp',
                    'spiral12': 'textures/spiral12.bmp',
                    'spiral14': 'textures/spiral14.bmp',
                    'tunnel0': 'textures/tunnel0.bmp',
                    'tunnel4': 'textures/tunnel4.bmp',
                    'tex01': 'textures/tex01.bmp',
                    'tex02': 'textures/tex02.bmp',
                    'tex04': 'textures/tex04.bmp',
                    'tex07': 'textures/tex07.bmp',
                    'tex08': 'textures/tex08.bmp',
                    'tex12': 'textures/tex12.bmp',
                    'tex13': 'textures/tex13.bmp',
                    'tex14': 'textures/tex14.bmp',
                    'tex16': 'textures/tex16.bmp',
                    'tex17': 'textures/tex17.bmp',
                    'tex18': 'textures/tex18.bmp',
                    'tex25': 'textures/tex25.bmp',
                    'tex26': 'textures/tex26.bmp',
                    'tex27': 'textures/tex27.bmp',
                    'tex29': 'textures/tex29.bmp',
                    'tex35': 'textures/tex35.bmp',
                    'tex36': 'textures/tex36.bmp',
                    'tex37': 'textures/tex37.bmp',
                    'tex38': 'textures/tex38.bmp',
                    'tex39': 'textures/tex39.bmp',
                    'tex43': 'textures/tex43.bmp',
                    'tex44': 'textures/tex44.bmp',
                    'tex59': 'textures/tex59.bmp',
                    'tex99': 'textures/tex99.bmp'})
        self.keystate = {K_UP: False, K_DOWN: False,
                         K_LEFT: False, K_RIGHT: False,
                         K_w: False, K_s: False,
                         K_a: False, K_d: False, # Interesting Keys
                         1: False, 3:False} # Mouse buttons

        if level:
            self.load (level)
        else:
            self.map.append (Section (Vector (0,0,0), None))

        self.hero = Hero(self.map[0], self)
        self.score = 0
        self.starting = 120
        self.leaving = 0
        self.leavingscale = 120.0

    def run(self):
        secs = self.map
        rate = 0

        clk = pygame.time.Clock()

        while not self.quit:
            #check for quit'n events
            self.handle_events()

            #clear screen and move camera
            glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)

            # Enemy movement, collisions
            here = self.hero.section.id
            i = 0
            while i < len(self.enemies):
                e = self.enemies[i]
                if here <= e.section.id <= here+1:
                    if (e.pos - self.hero.pos).norm() < (e.radius + self.hero.radius):
                        self.hero.crashWith(e)
                        e.hit('collision')
                if here-1 <= e.section.id <= here + self.MAX_SECTIONS:
                    e.move()
                if e.dead:
                    self.score += e.scoreValue
                    self.scoreLabels.append(Score(e.scoreValue, e.pos, e.section))
                    del self.enemies[i]
                else:
                    i += 1
            # Shot movement
            i = 0
            while i < len(self.shots):
                b = self.shots [i]
                b.move()
                b.check_hits(self)
                if b.dead:
                    del self.shots[i]
                else:
                    i += 1
            if not self.lost: self.hero.tick()
            self.hero.positionCamera()
            
            # Losing condition
            if self.hero.health <= 0:
                if self.leaving == 0:
                    effects.bigExplosion (self.hero.pos + self.hero.camera * 0.5, 0)
                    self.leaving = 120
                    self.lost = True
            
            # Winning condition: 4th from last section (haaack)
            if self.hero.section.next.next.next.next.next == self.hero.section.next.next.next.next:
                if self.leaving == 0:
                    self.leaving = 120
                    self.won = True
            
            # Fade out
            if self.leaving > 0:
                self.leaving -= 1
                if self.leaving == 0: self.quit = True
                k = self.leaving/self.leavingscale
                lightAmbient = [0.3*k, 0.3*k, 0.3*k, 1]
                lightDiffuse = [k, k, k, 1]
                glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient)
                glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse)
                
            self.drawScene()
            effects.tick(self)
            self.drawHud()
            pygame.display.flip()
            clk.tick(FPS)
            rate += 1
            if rate > 300:
                print "%d frames at %f FPS." % (rate, clk.get_fps())
                rate = 0

    def save (self, filename):
        """ Saves our current state to the specified file"""
        print 'Saving', filepath(filename)
        if os.path.exists(filepath(filename)): return False
        f = open (filepath(filename), 'w')
        f.write ('== Global ==\n')
        f.write ('Radius %f\n' % Section.defaultRadius)
        f.write ('NSides %d\n' % Section.sides)
        f.write ('== Points (floating-point triplets) ==\n')
        for s in self.map:
            f.write ("%f, %f, %f, %s\n" % (s.pos.x, s.pos.y, s.pos.z, s.texture.name))
        f.write ('== Enemies (type and parameters, one per line) ==\n')
        for e in self.enemies:
            f.write (e.serialize (self))
        f.write ('== End ==\n')
        return True

    def load (self, filename):
        def check(expr):
            if expr:
                print '.',
            else:
                print 'X',
        print 'Loading', filepath(filename),
        f = open (filepath(filename))
        line = f.readline()
        check (line == '== Global ==\n')
        diam = [x.strip() for x in f.readline().split()]
        check (diam[0] == 'Radius')
        Section.defaultRadius = float(diam[1])
        nsides = [x.strip() for x in f.readline().split()]
        check (nsides[0] == 'NSides')
        Section.sides = int(nsides[1])
        line = f.readline()
        check (line == '== Points (floating-point triplets) ==\n')
        line = f.readline()
        self.map = []
        last = None
        while not line.startswith('=='):
            nums = [float(num.strip()) for num in line.split(',')[:3]]
            last = Section (Vector (nums[0], nums[1], nums[2]), last)
            last.texture = self.mapTextures[line.split(',')[3].strip()]
            self.map.append(last)
            line = f.readline()
        self.map.append(LastSection(last))
        # Calculate coordinates and points before loading enemies!
        for s in self.map: s.updateCoords()
        for s in self.map: s.calculatePoints()

        check (line == '== Enemies (type and parameters, one per line) ==\n')
        line = f.readline()
        self.enemies = []
        enemies = {'Enemy1': lambda x: Enemy1(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['alpha1'], self, float(x[3])),
                   'SimpleEnemy': lambda x: SimpleEnemy(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['alpha1'], self, float(x[3])),
                   'Turret': lambda x: Turret(self.map[int(x[0])], float(x[1]), float(x[2]), float(x[3]), self.enemiesTextures['torreta'], self),
                   'EasyTracker': lambda x: Tracker(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['beta1'], self, reload=int(x[4])),
                   'Tracker': lambda x: Tracker(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['beta2'], self),
                   'Dodger': lambda x: Dodger(self.map[int(x[0])], self.enemiesTextures['gamma1'], self, speed=float(x[1]), reload=int(x[2])),
                   'Freighter': lambda x: Freighter(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['carguero'], self, float(x[3])),
                   'EnemyPasser': lambda x: EnemyPasser(self.map[int(x[0])], float(x[1]), float(x[2]), self.enemiesTextures['gamma1'], self, reload=int(x[3])),
                   'Alpha2': lambda x: Alpha(self.map[int(x[0])], self.enemiesTextures['alpha2'], self, float(x[1])),
                   # Broken
                   'Beta1': lambda x: Gamma(self.map[int(x[0])], self.enemiesTextures['beta1'], self),
                   # nekorB
                   'Gamma1': lambda x: Gamma(self.map[int(x[0])], self.enemiesTextures['gamma1'], self),
                   'Gamma2': lambda x: Gamma(self.map[int(x[0])], self.enemiesTextures['gamma2'], self),
                   'Delta1': lambda x: Delta(self.map[int(x[0])], self.enemiesTextures['delta1'], self),
                   'Delta2': lambda x: Delta(self.map[int(x[0])], self.enemiesTextures['delta2'], self),
                   'Epsilon1': lambda x: Epsilon(self.map[int(x[0])], self.enemiesTextures['epsilon1'], self),
                   'Epsilon2': lambda x: Epsilon(self.map[int(x[0])], self.enemiesTextures['epsilon2'], self),
                   }

        while not line.startswith('=='):
            enemy = [x.strip() for x in line.split(',')]
            if True:# Set to True for debugging!
                self.enemies.append(enemies[enemy[0]](enemy[1:]))
            else:
                try:
                    self.enemies.append(enemies[enemy[0]](enemy[1:]))
                    print '.',
                except:
                    print 'X',
            line = f.readline()
        check (line == '== End ==\n')
        print

    def handle_events (self):
        event = pygame.event.poll()
        while event.type != NOEVENT:
            if event.type == QUIT or (event.type == KEYDOWN and event.key == K_ESCAPE):
                if self.starting == 0 and self.leaving == 0:
                    self.leaving = 30
                    self.leavingscale = 30.0
            elif event.type == KEYDOWN and event.key == K_f:
                pygame.display.toggle_fullscreen()
            elif event.type == KEYDOWN and event.key in self.keystate:
                self.keystate [event.key] = True
            elif event.type == KEYUP and event.key in self.keystate:
                self.keystate [event.key] = False
            elif event.type == MOUSEMOTION:
                pygame.event.clear (MOUSEMOTION)
                self.mousePos = pygame.mouse.get_pos()
            elif event.type == MOUSEBUTTONDOWN and event.button in self.keystate:
                self.keystate [event.button] = True
            elif event.type == MOUSEBUTTONUP and event.button in self.keystate:
                self.keystate [event.button] = False
            elif event.type == KEYDOWN and event.key == K_p:
                self.hero.pause = not self.hero.pause
            elif event.type == KEYDOWN and event.key == K_o:
                # debug output. Useful for levelmaking
                print "Section: %d " % self.hero.section.id
            elif event.type == KEYDOWN and event.key == K_SPACE:
                self.hero.enableBurner()
            elif event.type == KEYUP and event.key == K_SPACE:
                self.hero.disableBurner()
            event = pygame.event.poll()
        if self.starting == 0 and self.leaving == 0:
            # Movement
            dx = dy = 0
            if self.keystate[K_DOWN] or self.keystate[K_s]: dy -= MOVE_SPEED
            if self.keystate[K_UP] or self.keystate[K_w]: dy += MOVE_SPEED
            if self.keystate[K_LEFT] or self.keystate[K_a]: dx += MOVE_SPEED
            if self.keystate[K_RIGHT] or self.keystate[K_d]: dx -= MOVE_SPEED
            self.hero.accel(dx, dy)
            if self.keystate[1]: # mouse button: fire
                self.hero.fire(self.mousePos)
            if self.keystate[3]: # mouse button: alternate fire
                self.hero.firePowerShot(self.mousePos)

    def drawScene(self):
        """ Draw the objects around your ship.  This function should leave all
            matrices untouched"""

        #glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glMatrixMode(GL_MODELVIEW)
        s = self.map[0]

        s = self.hero.section.prev
        if not s: s = self.hero.section
        drawn = 0
        while s is not s.next:
            s.draw()
            s = s.next
            drawn += 1
            if drawn >  self.MAX_SECTIONS: break

        for e in self.enemies:
            if self.hero.section.id-1 <= e.section.id < self.hero.section.id + self.MAX_SECTIONS:
                e.draw()
        for b in self.shots:
            b.draw()

        s = 0
        glBindTexture(GL_TEXTURE_2D, self.textures['gauge'].texid)
        while s < len(self.scoreLabels):
            if self.scoreLabels[s].draw(self):
                del self.scoreLabels[s]
            else:
                s += 1

    def drawHud(self):
        """ Draw the Head-Up Display.  This function should leave all
            matrices untouched"""
        # Cross-hair
        chs = 0.05
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        gluOrtho2D(0, 1, 1, 0)
        glColor3f(1, 1, 1)
        glDisable(GL_LIGHTING)
        glEnable (GL_BLEND)
        glBlendFunc (GL_SRC_COLOR, GL_ONE)
        glDisable(GL_DEPTH_TEST)
        glBindTexture (GL_TEXTURE_2D, self.alphaTextures['crosshair'].texid)
        glBegin(GL_QUADS)
        glTexCoord2f(0, 0)
        glVertex2f(self.mousePos[0] / self.width - chs, self.mousePos[1] / self.height - chs)
        glTexCoord2f(1, 0)
        glVertex2f(self.mousePos[0] / self.width + chs, self.mousePos[1] / self.height - chs)
        glTexCoord2f(1, 1)
        glVertex2f(self.mousePos[0] / self.width + chs, self.mousePos[1] / self.height + chs)
        glTexCoord2f(0, 1)
        glVertex2f(self.mousePos[0] / self.width - chs, self.mousePos[1] / self.height + chs)
        glEnd()
        # Life meter
        glBindTexture (GL_TEXTURE_2D, self.textures['gauge'].texid)
        hlth = self.hero.health / float(self.hero.fullHealth)
        if hlth < 0: hlth = 0
        glBegin(GL_QUADS)
        glTexCoord2f(0.01, 0.375)
        glVertex2f(0.84-0.15, 0.95)
        glTexCoord2f(0.5+hlth/2, 0.375)
        glVertex2f(0.84+0.15*hlth, 0.95)
        glTexCoord2f(0.5+hlth/2, 0.25)
        glVertex2f(0.84+0.15*hlth, 0.99)
        glTexCoord2f(0.01, 0.25)
        glVertex2f(0.84-0.15, 0.99)
        # Power meter
        power = self.hero.power / float(self.hero.fullPower)
        if power < 0: power = 0
        glTexCoord2f(0.01, 0.875)
        glVertex2f(0.01, 0.95)
        glTexCoord2f(0.5+power/2, 0.875)
        glVertex2f(0.16+0.15*power, 0.95)
        glTexCoord2f(0.5+power/2, 0.75)
        glVertex2f(0.16+0.15*power, 0.99)
        glTexCoord2f(0.01, 0.75)
        glVertex2f(0.01, 0.99)
        # Shield meter
        shield = self.hero.shield / float(self.hero.fullShield)
        if shield < 0: shield = 0
        glTexCoord2f(0.01, 0.625)
        glVertex2f(0.35, 0.95)
        glTexCoord2f(0.5+shield/2, 0.625)
        glVertex2f(0.5+0.15*shield, 0.95)
        glTexCoord2f(0.5+shield/2, 0.5)
        glVertex2f(0.5+0.15*shield, 0.99)
        glTexCoord2f(0.01, 0.5)
        glVertex2f(0.35, 0.99)
        glEnd()
        # Score
        glBegin(GL_QUADS)
        glTexCoord2f(0.00001, 0.25)
        glVertex2f(0.69, 0.01)
        glTexCoord2f(1, 0.25)
        glVertex2f(0.99, 0.01)
        glTexCoord2f(1, 0.125)
        glVertex2f(0.99, 0.05)
        glTexCoord2f(0.00001, 0.125)
        glVertex2f(0.69, 0.05)
        glEnd()
        self.drawScore(0.99, 0.03, 0, 0.04, 1.0, self.score)


        if self.starting > 0:
            self.starting -= 1
            k = (120-self.starting)/120.0
            lightAmbient = [0.5*k, 0.5*k, 0.5*k, 1]
            lightDiffuse = [k, k, k, 1]
            glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient)
            glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse)

            glColor4f(1, 1, 1, 1)
            glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            glBindTexture (GL_TEXTURE_2D, self.alphaTextures['ready'].texid)
            glBegin(GL_QUADS)
            glTexCoord2f(0, 0)
            glVertex2f(.4, .3)
            glTexCoord2f(1, 0)
            glVertex2f(.6, .3)
            glTexCoord2f(1, 1)
            glVertex2f(.6, .4)
            glTexCoord2f(0, 1)
            glVertex2f(.4, .4)
            glEnd()

        # Other effects
        effects.drawHud()
        # GLcleanUpTheMess
        glBlendFunc (GL_SRC_COLOR, GL_ONE)
        glEnable(GL_DEPTH_TEST)
        glEnable (GL_TEXTURE_2D)
        glEnable(GL_LIGHTING)
        glDisable(GL_BLEND)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)
        glPopMatrix()

    def duration(self, fps = 60):
        """How long (seconds) will the player take to move through this world, at given
        fps"""
        maplen = 0
        for s in self.map[:-1]:
            maplen += (s.pos - s.next.pos).norm()
        # maplen is in distance units. convert to ticks:
        maplen /= self.hero.speed
        # Now, convert ticks -> seconds:
        maplen /= fps
        return maplen

    def drawScore(self, x, y, z, h, align, score):
        if score < 0:
            glColor4f(1, 0, 0, 1)
        else:
            glColor4f(1, 1, 1, 1)
        msg = str(score)
        w = sum([chars[d][1] for d in msg])*h*8
        posx = x - w * align
        topy = y - h / 2.0
        boty = y + h / 2.0
        glBegin(GL_QUADS)
        for d in msg:
            glTexCoord2f(chars[d][0], 0.125)
            glVertex3f(posx, topy, z)
            glTexCoord2f(chars[d][0] + chars[d][1], 0.125)
            glVertex3f(posx+chars[d][1]*h*8, topy, z)
            glTexCoord2f(chars[d][0] + chars[d][1], 0)
            glVertex3f(posx+chars[d][1]*h*8, boty, z)
            glTexCoord2f(chars[d][0], 0)
            glVertex3f(posx, boty, z)
            posx += chars[d][1]*h*8
        glEnd()
